/*

*/

import tokenizer.*;
import tree.*;

/**
   Expression Language E2.

      E ::= 'a' '+' E   // right recursion means right associative
          | 'a'

   We can rewrite the grammar this way.

      E ::= 'a' [ '+' E ]

   This parser builds an abstract syntax tree
   for the expression being parsed. It will
   throw an error if the string cannot be parsed.

   Some example strings are:
      "a"
      "a + a"
      "a + a + a"
      "a + a + a + a"
*/
public class Parse2AST_E2
{
   public static Tree<Token> parseE2(String stringInE2)
   {
      // Tokenize the input string.
      final Tokenizer tokens = new Tokenizer(stringInE2);
      // Parse the list of tokens.
      final Tree<Token> ast = getE(tokens);
      // There shouldn't be any remaining tokens.
      if ( ! tokens.isEmpty() )
      {
         throw new ParseError("unexpected tokens, found " + tokens.peek());
      }
      return ast;
   }


   /**
      E ::= 'a' [ '+' E ]
   */
   private static Tree<Token> getE(Tokenizer tokens)
   {
      final Token tk1 = tokens.next();
      if ( ! tk1.lexeme.equals("a") )
      {
         throw new ParseError("expected \"a\", found " + tk1);
      }
      if ( ! tokens.isEmpty() && tokens.peek().lexeme.equals("+") )
      {
         final Token tk2 = tokens.next(); // consume the "+"
         final Tree<Token> ast = getE(tokens); // recursion
         return new LinkedTree<>(tk2,
                                 new LinkedTree<>(tk1),
                                 ast);
      }
      else
      {
         return new LinkedTree<>(tk1);
      }
   }


   public static void main(String[] args)
   {
      String[] expressions = {
                              "a",
                              "a + a",
                              "a + a + a",
                              "a + a + a + a",
                              "a+a+a+a",
                              "a+a+a-a",
                              "a + + a + a",
                              "a + a a + a",
                              "a + a +",
                              "+ a + a"
                             };

      int i = 0;
      for (final String str : expressions)
      {
         try
         {
            System.out.print("Parsing exp "+i+": \"" + str + "\"" );
            Tree<Token> ast = parseE2( str );
            System.out.println( " ... Parsed" );
            // Print the Tree using its toString() method.
            System.out.println( "AST = " + ast + "\n" );
            // Create dot and png files from the tree.
            final String baseName = String.format("Language_E2_Exp_%02d_AST", i);
            Tree2dot.tree2dot(ast, baseName);
            Tree2png.tree2png(baseName);
         }
         catch (TokenizeError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         catch (ParseError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         ++i;
      }
   }
}
