/*

*/

import tokenizer.*;
import tree.*;

/**
   Expression Language E6 but changed to be right associative.

      E ::= T '+' E    // right recursion means right associative
          | T

      T ::= '(' E ')'
          | 'a'

   We can rewrite the grammar this way.

      E ::= T [ '+' E ]
      T ::= '(' E ')'
          | 'a'

   This parser builds an abstract syntax tree
   for the expression being parsed. It will
   throw an error if the string cannot be parsed.

   Some example strings are:
      "(a + a)"
      "(a + a) + a"
      "a + (a + a)"
      "(a + a) + (a + a)"
      "a + (a + (a + a))"
      "((a + a) + a) + a"
      "a + (a + a) + a"
*/
public class Parse2AST_E6
{
   public static Tree<Token> parseE6(String stringInE6)
   {
      // Tokenize the input string.
      final Tokenizer tokens = new Tokenizer(stringInE6);
      // Parse the list of tokens.
      final Tree<Token> ast = getE(tokens);
      // There shouldn't be any remaining tokens.
      if ( ! tokens.isEmpty() )
      {
         throw new ParseError("unexpected tokens, found " + tokens.peek());
      }
      return ast;
   }


   /**
      E ::= T [ '+' E ]
   */
   private static Tree<Token> getE(Tokenizer tokens)
   {
      final Tree<Token> ast = getT(tokens);
      if ( ! tokens.isEmpty() && tokens.peek().lexeme.equals("+") )
      {
         final Token tk = tokens.next();  // consume the "+"
         final Tree<Token> subTree = getE(tokens); // recursion
         return new LinkedTree<>(tk, ast, subTree);
      }
      else
      {
         return ast;
      }
   }


   /**
      T ::= '(' E ')'
          | 'a'
   */
   private static Tree<Token> getT(Tokenizer tokens)
   {
      final Tree<Token> ast;
      if ( tokens.match("(") )
      {
         ast = getE(tokens);
         if ( ! tokens.match(")") )
         {
            throw new ParseError("expected \")\", found " + tokens.peek());
         }
      }
      else if ( tokens.peek().lexeme.equals("a") )
      {
         ast =  new LinkedTree<>(tokens.next());  // consume the "a"
      }
      else
      {
         throw new ParseError("expected \"a\", found " + tokens.peek());
      }
      return ast;
   }


   public static void main(String[] args)
   {
      String[] expressions = {
                              "(a + a)",
                              "(a + a) + a",
                              "a + (a + a)",
                              "(a + a) + (a + a)",
                              "a + (a + (a + a))",
                              "((a + a) + a) + a",
                              "a + (a + a) + a",
                              "a + ((a + a)) + a",
                              "a - (a + a) + a",
                              "a - (a + a) - a",
                              "a + (a + a)) + a",
                              "a + ((a + a) + a",
                             };
      int i = 0;
      for (final String str : expressions)
      {
         try
         {
            System.out.print("Parsing exp "+i+": \"" + str + "\"" );
            Tree<Token> ast = parseE6( str );
            System.out.println( " ... Parsed" );
            // Print the Tree using its toString() method.
            System.out.println( "AST = " + ast + "\n" );
            // Create dot and png files from the tree.
            final String baseName = String.format("Language_E6_Exp_%02d_AST", i);
            Tree2dot.tree2dot(ast, baseName);
            Tree2png.tree2png(baseName);
         }
         catch (TokenizeError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         catch (ParseError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         ++i;
      }
   }
}
