/*

*/

import tokenizer.*;
import tree.*;

/**
   Expression Language E8 but changed to be right associative.

      E ::= T '+' E    // right recursion means right associative
          | T

      T ::= 'a' '*' T  // right recursion means right associative
          | 'a'

   We can rewrite the grammar this way.

      E ::=  T  [ '+' E ]
      T ::= 'a' [ '*' T ]

   This parser builds an abstract syntax tree
   for the expression being parsed. It will
   throw an error if the string cannot be parsed.

   Some example strings are:
      "a + a"
      "a * a"
      "a + a * a"
      "a * a + a"
      "a + a * a + a"
      "a * a + a * a"
      "a * a * a * a"
      "a * a * a + a * a"
*/
public class Parse2AST_E8
{
   public static Tree<Token> parseE8(String stringInE8)
   {
      // Tokenize the input string.
      final Tokenizer tokens = new Tokenizer(stringInE8);
      // Parse the list of tokens.
      final Tree<Token> ast = getE(tokens);
      // There shouldn't be any remaining tokens.
      if ( ! tokens.isEmpty() )
      {
         throw new ParseError("unexpected tokens, found " + tokens.peek());
      }
      return ast;
   }


   /**
      E ::=  T  [ '+' E ]
   */
   private static Tree<Token> getE(Tokenizer tokens)
   {
      final Tree<Token> ast = getT(tokens);
      if ( ! tokens.isEmpty() && tokens.peek().lexeme.equals("+") )
      {
         final Token tk = tokens.next(); // consume the "+"
         final Tree<Token> subTree = getE(tokens); // recursion
         return new LinkedTree<>(tk, ast, subTree);
      }
      else
      {
         return ast;
      }
   }


   /**
      T ::= 'a' [ '*' T ]
   */
   private static Tree<Token> getT(Tokenizer tokens)
   {
      if ( ! tokens.peek().lexeme.equals("a") )
      {
         throw new ParseError("expected \"a\", found " + tokens.peek());
      }
      final Token tk1 = tokens.next(); // consume the "a"
      final Tree<Token> ast = new LinkedTree<>(tk1);
      if ( ! tokens.isEmpty() && tokens.peek().lexeme.equals("*") )
      {
         final Token tk2 = tokens.next();       // consume the "*"
         final Tree<Token> subTree = getT(tokens); // recursion
         return new LinkedTree<>(tk2, ast, subTree);
      }
      else
      {
         return ast;
      }
   }


   public static void main(String[] args)
   {
      String[] expressions = {
                              "a + a",
                              "a * a",
                              "a + a * a",
                              "a * a + a",
                              "a + a * a + a",
                              "a * a + a * a",
                              "a * a * a * a",
                              "a * a * a + a * a",
                              "a * + a",
                              "+ a * a",
                              "* a * a",
                              "a + + a",
                              "a * * a",
                              "a * a * ",
                              "a + a * ",
                              "a * a + "
                             };
      int i = 0;
      for (final String str : expressions)
      {
         try
         {
            System.out.print("Parsing exp "+i+": \"" + str + "\"" );
            Tree<Token> ast = parseE8( str );
            System.out.println( " ... Parsed" );
            // Print the Tree using its toString() method.
            System.out.println( "AST = " + ast + "\n" );
            // Create dot and png files from the tree.
            final String baseName = String.format("Language_E8_Exp_%02d_AST", i);
            Tree2dot.tree2dot(ast, baseName);
            Tree2png.tree2png(baseName);
         }
         catch (TokenizeError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         catch (ParseError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         ++i;
      }
   }
}
