/*

*/

import tokenizer.*;
import tree.*;

/**
   Expression Language E11 (compare with E3).

      E ::= E '+' T    // left recursion means left associative
          | E '-' T    // left recursion means left associative
          | T

      T ::= T '*' F    // left recursion means left associative
          | T '/' F    // left recursion means left associative
          | F

      F ::= '-' F      // right recursion means right associative
          | '+' F      // right recursion means right associative
          | '(' E ')'
          | N

      N ::= N D       // "N" for Number, "D for Digit
          | D

      D :: = '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'

   We can rewrite the grammar this way.

      E ::= T ( ( '+' | '-' ) T )*   // sum of Terms (iteration)
      T ::= F ( ( '*' | '/' ) F )*   // product of Factors (iteration)
      F ::= ( '+' | '-' ) F
          | '(' E ')'
          | N
      N ::= D ( D )*
      D :: = '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'

   This parser returns the value of the expression.
   It will throw an error if the string cannot be parsed.

   Some example strings are:
      "-5"
      "+5"
      "+(-5)"
      "-(+5)"
      "+-(-+5)"
      "(+2 + -3)"
      "-(4 + 3) / -2"
      "-2 * -(3 + 4)"
      "(2 + 3) * -(4 / 2)"
      "-10 / (1 + -(2 * -3))"
      "-(-(-2 + -3) * -4) + -5"
   This grammar can handle any valid four function
   arithmetic expression.
*/
public class Parse2Value_E11v2
{
   public static Tree<Token> parseE11(String stringInE11)
   {
      // Tokenize the input string.
      final Tokenizer tokens = new Tokenizer(stringInE11);
      // Parse the list of tokens.
      final Tree<Token> ast = getE(tokens);
      // There shouldn't be any remaining tokens.
      if ( ! tokens.isEmpty() )
      {
         throw new ParseError("unexpected tokens, found " + tokens.peek());
      }
      return ast;
   }


   /**
      E ::= T ( ( '+' | '-' ) T )*
   */
   private static Tree<Token> getE(Tokenizer tokens)
   {
      Tree<Token> ast = getT(tokens);

      while ( ! tokens.isEmpty() && ( tokens.peek().lexeme.equals("+")
                                   || tokens.peek().lexeme.equals("-") ) )
      {
         final Token tk = tokens.next(); // consume the operator
         final Tree<Token> subTree = getT(tokens); // recursion
         ast = new LinkedTree<>(tk,
                                ast,
                                subTree);
      }
      return ast;
   }


   /**
      T ::= F ( ( '*' | '/' ) F )*
   */
   private static Tree<Token> getT(Tokenizer tokens)
   {
      Tree<Token> ast = getF(tokens);

      while ( ! tokens.isEmpty() && ( tokens.peek().lexeme.equals("*")
                                   || tokens.peek().lexeme.equals("/") ) )
      {
         final Token tk = tokens.next(); // consume the operator
         final Tree<Token> subTree = getF(tokens); // recursion
         ast = new LinkedTree<>(tk,
                                ast,
                                subTree);
      }
      return ast;
   }


   /**
      F ::= ( '+' | '-' ) F
          | '(' E ')'
          | N
   */
   private static Tree<Token> getF(Tokenizer tokens)
   {
      final Tree<Token> ast;
      if ( tokens.peek().lexeme.equals("-")
        || tokens.peek().lexeme.equals("+") )
      {
         final Token tk = tokens.next(); // consume the operator
         final Tree<Token> subTree = getF(tokens);
         ast = new LinkedTree<>(tk, subTree);
      }
      else if ( tokens.match("(") )
      {
         ast = getE(tokens);
         if ( ! tokens.match(")") )
         {
            throw new ParseError("expected \")\", found " + tokens.peek());
         }
      }
      else
      {
         ast = getN(tokens);
      }
      return ast;
   }


   /**
      N ::= D ( D )*
   */
   private static Tree<Token> getN(Tokenizer tokens)
   {
      Tree<Token> result = getD(tokens);
      while ( ! tokens.isEmpty()
           && Character.isDigit(tokens.peek().lexeme.charAt(0)) )
      {
         Tree<Token> d = getD(tokens);
         result = new LinkedTree<>( new Token(
                                        result.root().lexeme + d.root().lexeme,
                                        result.root().position) );
      }
      return result;
   }


   /**
      D :: = '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'
   */
   private static Tree<Token> getD(Tokenizer tokens)
   {
      final Token tk = tokens.next();

      if ( ! Character.isDigit(tk.lexeme.charAt(0)) )
         throw new TokenizeError("expected digit, found " + tk);

      return new LinkedTree<>(tk);
   }


   public static void main(String[] args)
   {
      String[] expressions = {
                              "-51",
                              "+512",
                              "+(-50)",
                              "-(+5020)",
                              "+-(-+5432)",
                              "(+2 + -3)",
                              "-(4 + 3) / -2",
                              "-2 * -(3 + 4)",
                              "(2 + 3) * -(4 / 2)",
                              "-10 / (1 + -(2 * -3))",
                              "-(-(-2 + -3) * -4) + -5",
                              "-2 * -(-3 + -4) / +2",
                              "-10 / -(+(-1 - -2)) * -3",
                              "2 - (3 * 4) + 5",
                              "2 * (3 + 4) - 5",
                              "2 * (3 + 4) - -5",
                              "-(2 + 3 + 4) - 5",
                              "2 + -(3 * 4)) + 5",
                              "2 + -(-(3 * 4) + 5"
                             };
      int i = 0;
      for (final String str : expressions)
      {
         try
         {
            System.out.print("Parsing exp "+i+": \"" + str + "\"" );
            Tree<Token> ast = parseE11( str );
            System.out.println( " ... Parsed" );
            // Print the Tree using its toString() method.
            System.out.println( "AST = " + ast );
            // Evaluate the expression tree and print the result.
            final int r = Evaluate.eval(ast);
            System.out.println( r + " = " + str + "\n" );
            // Create dot and png files from the tree.
            final String baseName = String.format("Language_E11v2_Exp_%02d_AST", i);
            Tree2dot.tree2dot(ast, baseName);
            Tree2png.tree2png(baseName);
         }
         catch (TokenizeError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         catch (ParseError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         ++i;
      }
   }
}